dnl $1 = regex search prefix
dnl $2 = version string to parse
dnl $3 = name of var to set result
AC_DEFUN(GALEON_PARSE_VERSION,[
	_gpv_pfx=$1
	_gpv_vstr=$2
	# Solaris sed doesn't recognize \+, so use \{1,\} instead
	[_gpv_ver=`echo $_gpv_vstr | sed -n 's/^'${_gpv_pfx}'\([0-9]\{1,\}\)\.\([0-9]\{1,\}\)\.\([0-9]\{1,\}\).*$/\1*16777216+\2*65536+\3*256/p'`]
	if test -z "$_gpv_ver" ; then
		[_gpv_ver=`echo $_gpv_vstr | sed -n 's/^'${_gpv_pfx}'\([0-9]\{1,\}\)\.\([0-9]\{1,\}\).*$/\1*16777216+\2*65536/p'`]
	fi
	if test -n ["`echo $_gpv_vstr | sed -n 's/^'${_gpv_pfx}'[0-9.]\{1,\}+.*$/true/p'`"] ; then
		_gpv_ver="${_gpv_ver}+1"
	fi
	
	if test -z "$_gpv_ver" ; then 
		_gpv_ver="0"
	fi
	$3=`awk "BEGIN{ print $_gpv_ver; exit 0 }"`

])


dnl $1 = version int to parse
AC_DEFUN(GALEON_VERSION_RESULT,[
	# Solaris awk requires exit 0 to exit correctly instead of hanging
	_gvr_ver=$1
	[_gvr_ver_str=`awk "BEGIN{ a = int($_gvr_ver / 16777216); b = int($_gvr_ver / 65536 % 256); c = int($_gvr_ver / 256 % 256); d = int($_gvr_ver % 256); plus[0]=\"\"; plus[1]=\"+\"; printf \"%d.%d.%d%s\n\", a, b, c, plus[d]; exit 0 }"`]
	AC_MSG_RESULT($_gvr_ver_str)
])


dnl $1 = version required
dnl $2 = action if required version is found
dnl $3 = action if required version is not found
dnl prerequisites: MOZILLA_HOME
dnl sets: MOZILLA_VERSION
AC_DEFUN(MOZILLA_CHECK_VERSION,[
	_moz_prefs_file=$MOZILLA_HOME/defaults/pref/all.js

	if test -f "$_moz_prefs_file" ; then
		_user_agent=`grep useragent.misc $_moz_prefs_file`
	else
		AC_MSG_WARN([couldn't find file: $_moz_prefs_file])
		_user_agent=unknown
	fi

	GALEON_PARSE_VERSION("", $1, _moz_req_version)
	GALEON_PARSE_VERSION(".*rv:", $_user_agent, MOZILLA_VERSION)
	
	if test "$MOZILLA_VERSION" -lt "$_moz_req_version" ; then
		ifelse([$3], , , [$3])
	else
		ifelse([$2], , :, [$2])
	fi
])


dnl $1 = required version
dnl sets: MOZILLA_LIB_DIR, MOZILLA_INCLUDE_DIR, MOZILLA_HOME, MOZILLA_COMPONENTS_DIR, MOZILLA_VERSION
AC_DEFUN(MOZILLA_INIT, [
	AC_MSG_CHECKING(for Mozilla)

	MOZILLA_LIB_DIR=/usr/lib/mozilla
	MOZILLA_INCLUDE_DIR=/usr/include/mozilla

	AC_ARG_WITH(mozilla-libs,
		[  --with-mozilla-libs     Set mozilla 'lib' place ],
		[ MOZILLA_LIB_DIR="$withval" ])

	AC_ARG_WITH(mozilla-includes,
		[  --with-mozilla-includes Set mozilla 'include' place ],
		[ MOZILLA_INCLUDE_DIR="$withval" ])

	AC_ARG_WITH(mozilla-home,
		[  --with-mozilla-home     Set mozilla 'home' place ],
		[ MOZILLA_HOME="$withval" ],
		[ MOZILLA_HOME="$MOZILLA_LIB_DIR" ])

	AC_ARG_WITH(mozilla-components,
		[  --with-mozilla-components  Set mozilla 'components' place ],
		[ MOZILLA_COMPONENTS_DIR="$withval" ],
		[ MOZILLA_COMPONENTS_DIR="$MOZILLA_HOME/components" ])

	AC_MSG_RESULT($MOZILLA_HOME)

	AC_MSG_CHECKING(Mozilla version)

	MOZILLA_CHECK_VERSION($1, ,
	[
		AC_MSG_ERROR(
		[
			*** Mozilla $1 is required 
			*** A package for $1 is available here:
			*** http://ftp.mozilla.org/pub/mozilla/releases/mozilla$1/.
		])
	])

	GALEON_VERSION_RESULT($MOZILLA_VERSION)
])


dnl prerequisites: MOZILLA_INCLUDE_DIR
dnl sets: NSPR_INCLUDE_DIR
AC_DEFUN(NSPR_HEADERS,[
	AC_ARG_WITH(nspr-includes,
		[  --with-nspr-includes    Set nspr 'include' place ],
		[ NSPR_INCLUDE_DIR="$withval" ],
		[ 
			NSPR_INCLUDE_DIR="$MOZILLA_INCLUDE_DIR/nspr"
			if test ! -d "$NSPR_INCLUDE_DIR"; then
				NSPR_INCLUDE_DIR="/usr/include/nspr"
			fi
	])
])
