/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#ifndef __TCLXPCOMPRIVATE_H__
#define __TCLXPCOMPRIVATE_H__

#ifndef MOZILLA_STRICT_API
#define MOZILLA_STRICT_API
#endif

// Tcl_GetIndexFromObj requires no const in 8.3 and const in 8.4
#define USE_NON_CONST

#include "tclxpcom.h"
#include "xptcall.h"
#include "nsISupports.h"
#include "nsISupportsUtils.h"
#include "nsIClassInfo.h"
#include "nsIInterfaceInfo.h"
#include "nsIServiceManager.h"
#include "nsIInterfaceInfoManager.h"
#include "nsIComponentRegistrar.h"
#include "nsIMemory.h"
#include "nsWeakReference.h"

extern Tcl_Encoding ascii_enc;


#ifdef TCL_WIDE_INT_TYPE
#   define USE_TCL_WIDEINT
#endif


#ifdef USE_TCL_WIDEINT

typedef Tcl_WideInt TclXPCOM_WideInt;
#define TclXPCOM_NewWideIntObj     Tcl_NewWideIntObj
#define TclXPCOM_GetWideIntFromObj Tcl_GetWideIntFromObj

#else /* !defined(USE_TCL_WIDEINT) */

typedef PRInt64 TclXPCOM_WideInt;
Tcl_Obj *TclXPCOM_NewWideIntObj     (TclXPCOM_WideInt wideint);
int      TclXPCOM_GetWideIntFromObj (Tcl_Interp *interp, Tcl_Obj *objPtr, TclXPCOM_WideInt *wideint);

#endif


extern nsIServiceManager       *servmanager;
extern nsIInterfaceInfoManager *manager;
extern nsIComponentRegistrar   *registrar;
extern nsIMemory               *memmgr;

extern nsIID nsisupports_iid;
extern nsIID nsiclassinfo_iid;
extern nsIID nsiservicemanager_iid;

extern nsIInterfaceInfo *nsisupports_info;
extern nsIInterfaceInfo *nsiservicemanager_info;
extern nsIInterfaceInfo *nsiclassinfo_info;

extern Tcl_ObjType ISupportsType;

/**
 *  Flags for tclparams
 */
#define PARAM_IS_SKIPPED         0x0100
#define PARAM_IS_RETVAL          0x0200
#define PARAM_IS_SKIPPED_IN      0x0400
#define PARAM_IS_SKIPPED_OUT     0x0800
#define PARAM_IS_SHARED_IN       0x1000
#define PARAM_IS_SHARED_OUT      0x2000
#define PARAM_INDEX_MASK         0x00ff


/**
 *  
 */
typedef struct MethodInfo
{
    PRUint8  paramcount;             ///< number of native parameters
    PRUint8  tclparamcount;          ///< number of tcl parameters
    PRUint16 *tclparams;             ///< index of tcl parameters
    nsXPTParamInfo *paraminfos;      ///< parameter info
} MethodInfo;


/**
 *  
 */
typedef struct IdentifierInfo_
{
    char *name;
    PRUint8 type;
    union
    {
        PRUint16 methodindex;
        struct
        {
            PRUint16 getterindex;
            PRUint16 setterindex;
        } attributeindices;
    } indices;
} IdentifierInfo;


/**
 *  
 */
typedef struct InterfaceInfo
{
//  int timestamp;                  ///< time of last access
    nsIInterfaceInfo *info;
    nsIInterfaceInfo *parentinfo;
//  const nsIID *iid;

    PRUint16 imethodcount;
    PRUint16 methodcount;           ///< number of methods
    PRUint16 attributecount;        ///< number of attributes
    PRUint16 constantcount;         ///< number of constants

    PRUint16 baseimethodindex;
    PRUint16 basemethodindex;
    PRUint16 baseattributeindex;
    PRUint16 baseconstantindex;

    MethodInfo     *imethods;       ///< cache of interface methods
    IdentifierInfo *methods;        ///< cache of method info
    IdentifierInfo *attributes;     ///< cache of attribute info
    char          **constantnames;  ///< constant names
    Tcl_Obj       **constantvalues; ///< cache of constant values
} InterfaceInfo;


/**
 *
 */
typedef struct ClassInfoSignature
{
    int                infocount;   ///< number of infos in list
    nsIInterfaceInfo **infolist;    ///< sorted list of infos
} ClassInfoSignature;


/**
 *
 */
typedef struct ClassIdentifierInfo
{
    IdentifierInfo idinfo;
    nsIInterfaceInfo *info;         ///< info
    PRUint8 qualified;              ///< non-zero if this identifier is qualified
} ClassIdentifierInfo;


/**
 *
 */
typedef struct ClassInfo_
{
//  int timestamp;                  ///< time of last access
    ClassInfoSignature  signature;  ///< signature of class
    PRUint16            methodcount;///< number of methods
    PRUint16         attributecount;///< number of attributes
    ClassIdentifierInfo *methods;   ///< cache of method info
    ClassIdentifierInfo *attributes;///< cache of attribute info
} ClassInfo;


/**
 *
 */
typedef struct InterfaceRef
{
    int refcount;                ///< number of Tcl_Objs that point to this
    Tcl_HashTable    refhash;    ///< mapping of interface infos to interface pointers
    nsIClassInfo     *cinforef;  ///< pointer to the nsIClassInfo for this ref
    ClassInfo        *cinfo;     ///< classinfo
    int              cisupported;///< 1 if classinfo is supported for this ref
} InterfaceRef;


/**
 *  
 */
class TclXPCOMStub : public nsXPTCStubBase,
                     public nsSupportsWeakReference
{
    public:

    nsIInterfaceInfo         *info;     ///< pointer to the Interface information
    struct TclXPCOMComponent *component;///< pointer to the component information
    Tcl_Obj                 **bindings; ///< script bindings to interface method

    TclXPCOMStub  (nsIInterfaceInfo *interfaceinfo);
    ~TclXPCOMStub (void);

    NS_IMETHOD QueryInterface(REFNSIID aIID, void** aInstancePtr);
    NS_IMETHOD_(nsrefcnt) AddRef(void);
    NS_IMETHOD_(nsrefcnt) Release(void);
    NS_IMETHOD GetInterfaceInfo(nsIInterfaceInfo** info);
    NS_IMETHOD CallMethod(PRUint16 methodIndex, const nsXPTMethodInfo *info, nsXPTCMiniVariant *params);

    void     SetBinding   (PRUint16 methodindex, Tcl_Obj *script);
    void     UnsetBinding (PRUint16 methodindex);
    Tcl_Obj *GetBinding   (PRUint16 methodindex);
};


/**
 *
 */
typedef struct TclXPCOMComponent
{
    nsrefcnt       refcount;  ///< our refcount
    Tcl_Interp    *interp;    ///< the interp where we were created
    Tcl_ThreadId   thread;    ///< the thread where we were created
    ClassInfo     *cinfo;     ///< class info for this component
    TclXPCOMStub **stubs;     ///< interface stubs that are implemented by this component
    Tcl_Obj       *destructor;///< script called when component is deleted
    int            nesting;   ///< set to 1 if we're running a script for QI, AddRef, or Release
    int            nestlevel; ///< current nesting level
} TclXPCOMComponent;


void TclXPCOM_InitInterfaceRef  (void);
void TclXPCOM_InitClassInfo     (void);
void TclXPCOM_InitInterfaceInfo (void);
void TclXPCOM_RegisterObjTypes  (void);

int           AddRefInterface    (nsISupports *isupports, nsIInterfaceInfo *interfaceinfo, Tcl_HashEntry **entry);
int           ReleaseInterface   (nsISupports *isupports);
int           QueryInterfaceIRef (Tcl_Interp *interp, InterfaceRef *iref, nsIInterfaceInfo *queryinfo, nsISupports **result);
int           GetMethodFromIRef  (Tcl_Interp *interp, nsISupports *isupports, nsIInterfaceInfo *info, Tcl_Obj *identifier, int type, nsIInterfaceInfo **interfaceinfo, PRUint16 *methodindex);
InterfaceRef *GetInterfaceRef    (nsISupports *isupports);

TclXPCOMComponent *NewTclXPCOMComponent  (Tcl_Interp *interp, Tcl_ThreadId thread, ClassInfo *cinfo);
void               FreeTclXPCOMComponent (TclXPCOMComponent *component);
int                BindScript            (Tcl_Interp *interp, TclXPCOMComponent *component, Tcl_Obj *identifier, Tcl_Obj *script, int flags);

InterfaceInfo *GetInterfaceInfo (nsIInterfaceInfo *interfaceinfo);
MethodInfo    *GetMethodInfo    (nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex);

int IsInInterface (nsIInterfaceInfo *infoa, nsIInterfaceInfo *infob);
int isInInterface (nsIInterfaceInfo *interfaceinfo, nsIID *iid);

int      Tcl2Native   (Tcl_Interp *interp, Tcl_Obj *value, void **params, void *param, nsXPTType *type, int shared, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex, nsXPTParamInfo *paraminfo);
nsresult Native2Tcl   (void **params, void *param, nsXPTType *type, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex, nsXPTParamInfo *paraminfo, Tcl_Obj **result);
void     CleanupParam (void **params, void *param, nsXPTType *type, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex, nsXPTParamInfo *paraminfo);

nsresult TclXPCOM_CallMethod (Tcl_Interp *interp, nsXPTCMiniVariant *variants, nsISupports *isupports, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex, Tcl_Obj *script);

void *CloneMemory (void *ptr, size_t size);

#endif /* __TCLXPCOMPRIVATE_H__ */
