/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TclXPCOM.
 * 
 * The Initial Developer of the Original Code is Mark Follett.
 * Portions created by Mark Follett are Copyright (C) 2001-2002
 * Mark Follett.  All Rights Reserved.
 * 
 * Contributor(s):
 *     Mark Follett <mef123@myrealbox.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#ifndef __TCLXPCOM_H__
#define __TCLXPCOM_H__

#include "tcl.h"
#include "nsISupports.h"
#include "nsIInterfaceInfo.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * if the BUILD_TCLXPCOM macro is defined, the assumption is that we are
 * building the dynamic library.
 */
#ifdef BUILD_TCLXPCOM
#  undef TCL_STORAGE_CLASS
#  define TCL_STORAGE_CLASS DLLEXPORT
#endif

#define TCLXPCOM_MOZHOMEVARNAME    "::xpcom::mozilla_home"
#define TCLXPCOM_LASTRESULTVARNAME "::xpcom::lastResult"
#define TCLXPCOM_ERRORMSGPROC      "::xpcom::errormsg"
#define TCLXPCOM_BGERRORPROC       "::xpcom::bgerror"

EXTERN int Tclxpcom_Init _ANSI_ARGS_((Tcl_Interp *interp));
EXTERN int Tclxpcom_SafeInit _ANSI_ARGS_((Tcl_Interp *interp));


// custom tcl object types
#define TCLXPCOM_ISUPPORTSTYPENAME  "TclXPCOM_Interface_Reference"
#define TCLXPCOM_NSIDTYPENAME       "TclXPCOM_Unique_ID"
#define TCLXPCOM_POINTERTYPENAME    "TclXPCOM_Void_Pointer"
#define TCLXPCOM_IDENTIFIERTYPENAME "TclXPCOM_Identifier"

EXTERN Tcl_Obj *TclXPCOM_NewISupportsObj  (nsISupports *isupports, nsIInterfaceInfo *interfaceinfo);
EXTERN Tcl_Obj *TclXPCOM_NewNsIDObj       (nsID *id);
EXTERN Tcl_Obj *TclXPCOM_NewPointerObj    (void *ptr);
EXTERN Tcl_Obj *TclXPCOM_NewIdentifierObj (nsIInterfaceInfo *interfaceinfo, char *identifier);

EXTERN int TclXPCOM_GetISupportsFromObj  (Tcl_Interp *interp, Tcl_Obj *objPtr, nsISupports **ref, nsIInterfaceInfo **info);
EXTERN int TclXPCOM_QueryISupportsObj    (Tcl_Interp *interp, Tcl_Obj *objPtr, nsIInterfaceInfo *info, nsISupports **ref);
EXTERN int TclXPCOM_GetNsIDFromObj       (Tcl_Interp *interp, Tcl_Obj *objPtr, nsID **id);
EXTERN int TclXPCOM_GetPointerFromObj    (Tcl_Interp *interp, Tcl_Obj *objPtr, void **ptr);
EXTERN int TclXPCOM_GetIdentifierFromObj (Tcl_Interp *interp, Tcl_Obj *objPtr, nsIInterfaceInfo **interfaceinfo, char **identifier);

EXTERN int TclXPCOM_IsNULL (Tcl_Obj *objPtr);


// flags for invoke and stubs facility
#define FORCE_HIDDEN         0x01
#define FORCE_NOTXPCOM       0x02
#define FORCE_NOTSCRIPTABLE  0x04


EXTERN void TclXPCOM_RegisterCommands (Tcl_Interp *interp);

EXTERN Tcl_ObjCmdProc TclXPCOM_InvokeObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_WrapObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_ThisObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_ConstantObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_InfoObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_NullObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_IsNullObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_PointerObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_AddressObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_FromStringObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_SetReturnCodeObjCmd;
EXTERN Tcl_ObjCmdProc TclXPCOM_IdObjCmd;

// invoke facility
EXTERN int TclXPCOM_Invoke (Tcl_Interp *interp, Tcl_Obj *CONST objv[], nsISupports *isupports, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex);
// int TclXPCOM_Invoke (Tcl_Interp *interp, Tcl_Obj *CONST objv[], nsISupports *isupports, TclXPCOM_MethodInfo methodinfo);

// stubs facility
EXTERN void     TclXPCOM_GetCurrentComponentContext (nsIInterfaceInfo **interfaceinfo, nsISupports **isupports);
EXTERN int      TclXPCOM_SetReturnCode              (nsresult res);

EXTERN int TclXPCOM_CheckInterfaceIsReflectable (Tcl_Interp *interp, nsIInterfaceInfo *interfaceinfo, int flags);
EXTERN int TclXPCOM_CheckMethodIsReflectable    (Tcl_Interp *interp, nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex, int flags);


typedef struct IdentifierInfo_ IdentifierInfo;
typedef struct ClassInfo_      ClassInfo;

// interfaceinfo
#define INTERFACEINFO_METHOD  0x01
#define INTERFACEINFO_GETTER  0x02
#define INTERFACEINFO_SETTER  0x04

EXTERN int        TclXPCOM_GetIdentifierInfoFromInterface (Tcl_Interp *interp, nsIInterfaceInfo *info, Tcl_Obj *identifier, int type, IdentifierInfo **idinfo);
EXTERN int        TclXPCOM_GetIdentifierInfoFromClass     (Tcl_Interp *interp, ClassInfo *cinfo, Tcl_Obj *identifier, int type, nsIInterfaceInfo **info, IdentifierInfo **idinfo);
EXTERN int        TclXPCOM_GetMethodIndex                 (Tcl_Interp *interp, IdentifierInfo *idinfo, int type, PRUint16 *methodindex);
EXTERN int        TclXPCOM_GetConstant                    (Tcl_Interp *interp, nsIInterfaceInfo *interfaceinfo, Tcl_Obj *constantname, Tcl_Obj **result);
EXTERN ClassInfo *TclXPCOM_GetClassInfo                   (int iidcount, nsIID **iids);



//EXTERN TclXPCOMInterface TclXPCOM_GetInterface      (nsIInterfaceInfo *interfaceinfo);
//EXTERN TclXPCOMClass     TclXPCOM_GetClass          (int iidcount, nsIID **iids);
//EXTERN TclXPCOMMethod    TclXPCOM_GetMethod         (nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex);

//EXTERN Tcl_Obj          *TclXPCOM_GetInterfaceConstantValue  (TclXPCOMInterface interface, int constantindex);

//EXTERN int               TclXPCOM_GetMethodFromInterface (Tcl_Interp *interp, TclXPCOMInterface interface, Tcl_Obj *identifier, int type, TclXPCOMMethod *method);
//EXTERN int               TclXPCOM_GetMethodFromClass     (Tcl_Interp *interp, TclXPCOMClass class, Tcl_Obj *identifier, int type, nsIInterfaceInfo **info, MethodInfo **minfo);
//EXTERN int               TclXPCOM_GetMethod              (Tcl_Interp *interp, TclXPCOMInterface interface, MethodInfo *minfo, int type, TclXPCOMMethod *method);
//EXTERN int               TclXPCOM_GetConstant            (Tcl_Interp *interp, TclXPCOMInterface interface, Tcl_Obj *constantname, Tcl_Obj **result);


EXTERN void TclXPCOM_SetErrorCode    (Tcl_Interp *interp, nsresult res);
//returns 1 if successful,  0 if failure
EXTERN int  TclXPCOM_GetErrorCode    (Tcl_Interp *interp, nsresult *res);
EXTERN void TclXPCOM_BackgroundError (Tcl_Interp *interp);

EXTERN void TclXPCOM_SetErrorMessage (Tcl_Interp *interp, nsresult res);
EXTERN void TclXPCOM_SetError        (Tcl_Interp *interp, nsresult res);


// utility functions
EXTERN Tcl_Obj *TclXPCOM_ListInterfaceMethods    (nsIInterfaceInfo *interfaceinfo);
EXTERN Tcl_Obj *TclXPCOM_ListInterfaceAttributes (nsIInterfaceInfo *interfaceinfo);
EXTERN Tcl_Obj *TclXPCOM_ListInterfaceConstants  (nsIInterfaceInfo *interfaceinfo);
EXTERN Tcl_Obj *TclXPCOM_ListClassMethods        (ClassInfo *cinfo);
EXTERN Tcl_Obj *TclXPCOM_ListClassAttributes     (ClassInfo *cinfo);

EXTERN Tcl_Obj *TclXPCOM_ListMethodArguments     (nsIInterfaceInfo *interfaceinfo, PRUint16 methodindex);
EXTERN Tcl_Obj *TclXPCOM_GetAttributeType        (nsIInterfaceInfo *interfaceinfo, IdentifierInfo *idinfo);
EXTERN Tcl_Obj *TclXPCOM_GetReturnValueType      (nsIInterfaceInfo *interfaceinfo, IdentifierInfo *idinfo);


# undef TCL_STORAGE_CLASS
# define TCL_STORAGE_CLASS DLLIMPORT
    
#ifdef __cplusplus
}
#endif

#endif /* __TCLXPCOM_H__ */
