# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is TclXPCOM.
# 
# The Initial Developer of the Original Code is Mark Follett.
# Portions created by Mark Follett are Copyright (C) 2001-2002
# Mark Follett.  All Rights Reserved.
# 
# Contributor(s):
#     Mark Follett <mef123@myrealbox.com> (Original Author)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.

package provide mozilla 1.0

namespace eval ::mozilla {
  variable libprefix
  variable versions
  variable packagedir
  global tcl_platform

  set packagedir [file dirname [file join [pwd] [info script]]]

  set libprefix ""
  if {$tcl_platform(platform) == "unix"} {
    set libprefix "lib"
  }

  proc normalizeVersion {string} {
    regexp {^([0-9]+)?[.]?([0-9]+)?[.]?([0-9]+)?} $string x major minor patch
    if {$major == ""} {
      set result 0
    } else {
      set result $major
    }
    if {$minor == ""} {
      set minor 0
    }
    append result ".$minor"
    if {$patch != ""} {
      append result ".$patch"
    }
    return $result
  }

  proc compareVersions {versiona versionb} {
    array set priority {
      "PACKAGEDIR" 4
      "ENVVAR"     3
      "SEAMONKEY"  2
      "GRE"        1
    }
    set result [package vcompare [lindex $versiona 0] [lindex $versionb 0]]
    if {$result != 0} {return $result}
    return [string compare $priority([lindex $versiona 2]) $priority([lindex $versionb 2])]
  }

  proc mozExists {home} {
    variable libprefix
    append libname $libprefix "xpcom" [info sharedlibextension]
    file exists [file join $home $libname]
  }

  proc getMozVersion {home} {
    set version ""
    set preffile [file join $home defaults pref all.js]
    if {[file exists $preffile]} {
      set ofile [open $preffile r]
      while {![eof $ofile]} {
        set line [gets $ofile]
        if {[regexp {useragent[.]misc.*rv:([^"]+)} $line x version]} break
      }
      close $ofile
    }
    return [normalizeVersion $version]
  }

  proc keyExists {keyname} {
    expr {![catch {registry keys $keyname}]}
  }

  proc getMozInstallations {} {
    global tcl_platform
    global env
    variable packagedir

    set versions {}

    if {[string equal $tcl_platform(platform) "windows"] && ![catch {package require registry}]} {

      set MOZILLA_ROOT_KEY "HKEY_LOCAL_MACHINE\\Software\\mozilla.org"
      set MOZILLA_GRE_KEY  "$MOZILLA_ROOT_KEY\\GRE"
      set MOZILLA_MOZ_KEY  "$MOZILLA_ROOT_KEY\\Mozilla"

      ### Check for Mozilla Gecko Runtime Environment Installations
      if {[keyExists $MOZILLA_GRE_KEY]} {
        foreach subkey [registry keys $MOZILLA_GRE_KEY] {
          set home [registry get "$MOZILLA_GRE_KEY\\$subkey" GreHome]
          if {[mozExists $home]} {
            lappend versions [list [normalizeVersion $subkey] $home "GRE"]
          }
        }
      }

      ### Check for Mozilla Seamonkey Installations
      if {[keyExists $MOZILLA_MOZ_KEY]} {
        foreach subkey [registry keys $MOZILLA_MOZ_KEY] {
          set home [registry get "$MOZILLA_MOZ_KEY\\$subkey\\Main" "Install Directory"]
          if {[mozExists $home]} {
            lappend versions [list [normalizeVersion $subkey] $home "SEAMONKEY"]
          }
        }
      }
    }

    ### Check for presence of the MOZILLA_FIVE_HOME environment variable
    if {[info exists env(MOZILLA_FIVE_HOME)]} {
      set home $env(MOZILLA_FIVE_HOME)
      if {[mozExists $home]} {
        lappend versions [list [getMozVersion $home] $home "ENVVAR"]
      }
    }

    ### Check if Mozilla is in the package directory
    if {[mozExists $packagedir]} {
      lappend versions [list [getMozVersion $packagedir] $packagedir "PACKAGEDIR"]
    }

    return [lsort -decreasing -command compareVersions $versions]
  }

  proc getHome {{reqversion ""}} {
    variable versions
    if {$reqversion != ""} {
      set reqversion [normalizeVersion $reqversion]
      foreach version $versions {
        if {[package vsatisfies [lindex $version 0] $reqversion]} {
          set found $version
          break
        }
      }
      if {![info exists found]} {
        error "can't find mozilla $reqversion"
      }
    } else {
      if {[llength $versions] > 0} {
        set found [lindex $versions 0]
      } else {
        error "can't find mozilla"
      }
    }
    return [lindex $found 1]
  }

  proc getVersions {} {
    variable versions
    return $versions
  }

  set versions [getMozInstallations]
}
