# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is TclXPCOM.
# 
# The Initial Developer of the Original Code is Mark Follett.
# Portions created by Mark Follett are Copyright (C) 2001-2002
# Mark Follett.  All Rights Reserved.
# 
# Contributor(s):
#     Mark Follett <mef123@myrealbox.com> (Original Author)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.

package provide tclxpcom 0.1

namespace eval xpcom {

  variable lastResult

  variable errormsg
  variable modulenames

  variable smanager
  variable compmanager
  variable infomanager

  variable NS_ERROR_MODULE_BASE_OFFSET             [expr 0x45]
  variable NS_ERROR_MODULE_XPCOM                   1
  variable NS_ERROR_MODULE_BASE                    2
  variable NS_ERROR_MODULE_GFX                     3
  variable NS_ERROR_MODULE_WIDGET                  4
  variable NS_ERROR_MODULE_CALENDAR                5
  variable NS_ERROR_MODULE_NETWORK                 6
  variable NS_ERROR_MODULE_PLUGINS                 7
  variable NS_ERROR_MODULE_LAYOUT                  8
  variable NS_ERROR_MODULE_HTMLPARSER              9
  variable NS_ERROR_MODULE_RDF                     10
  variable NS_ERROR_MODULE_UCONV                   11
  variable NS_ERROR_MODULE_REG                     12
  variable NS_ERROR_MODULE_FILES                   13
  variable NS_ERROR_MODULE_DOM                     14
  variable NS_ERROR_MODULE_IMGLIB                  15
  variable NS_ERROR_MODULE_MAILNEWS                16
  variable NS_ERROR_MODULE_EDITOR                  17
  variable NS_ERROR_MODULE_XPCONNECT               18
  variable NS_ERROR_MODULE_PROFILE                 19
  variable NS_ERROR_MODULE_LDAP                    20
  variable NS_ERROR_MODULE_GENERAL                 51

  variable NS_OK                                   [expr 0]
  variable NS_COMFALSE                             [expr 1]
  variable NS_ERROR_BASE                           [expr 0xC1F30000]
  variable NS_ERROR_NOT_INITIALIZED                [expr $NS_ERROR_BASE + 1]
  variable NS_ERROR_ALREADY_INITIALIZED            [expr $NS_ERROR_BASE + 2]
  variable NS_ERROR_NOT_IMPLEMENTED                [expr 0x80004001]
  variable NS_NOINTERFACE                          [expr 0x80004002]
  variable NS_ERROR_NO_INTERFACE                   [expr $NS_NOINTERFACE]
  variable NS_ERROR_INVALID_POINTER                [expr 0x80004003]
  variable NS_ERROR_NULL_POINTER                   [expr $NS_ERROR_INVALID_POINTER]
  variable NS_ERROR_ABORT                          [expr 0x80004004]
  variable NS_ERROR_FAILURE                        [expr 0x80004005]
  variable NS_ERROR_UNEXPECTED                     [expr 0x8000ffff]
  variable NS_ERROR_OUT_OF_MEMORY                  [expr 0x8007000e]
  variable NS_ERROR_ILLEGAL_VALUE                  [expr 0x80070057]
  variable NS_ERROR_INVALID_ARG                    [expr $NS_ERROR_ILLEGAL_VALUE]
  variable NS_ERROR_NO_AGGREGATION                 [expr 0x80040110]
  variable NS_ERROR_NOT_AVAILABLE                  [expr 0x80040111]
  variable NS_ERROR_FACTORY_NOT_REGISTERED         [expr 0x80040154]
  variable NS_ERROR_FACTORY_REGISTER_AGAIN         [expr 0x80040155]
  variable NS_ERROR_FACTORY_NOT_LOADED             [expr 0x800401f8]
  variable NS_ERROR_FACTORY_NO_SIGNATURE_SUPPORT   [expr $NS_ERROR_BASE + 0x101]
  variable NS_ERROR_FACTORY_EXISTS                 [expr $NS_ERROR_BASE + 0x100]
  variable NS_ERROR_PROXY_INVALID_IN_PARAMETER     [expr 0x80010010]
  variable NS_ERROR_PROXY_INVALID_OUT_PARAMETER    [expr 0x80010011]

  array set modulenames [list \
    $NS_ERROR_MODULE_XPCOM       "xpcom"       \
    $NS_ERROR_MODULE_BASE        "base"        \
    $NS_ERROR_MODULE_GFX         "graphics"    \
    $NS_ERROR_MODULE_WIDGET      "widget"      \
    $NS_ERROR_MODULE_CALENDAR    "calendar"    \
    $NS_ERROR_MODULE_NETWORK     "network"     \
    $NS_ERROR_MODULE_PLUGINS     "plugins"     \
    $NS_ERROR_MODULE_LAYOUT      "layout"      \
    $NS_ERROR_MODULE_HTMLPARSER  "htmlparser"  \
    $NS_ERROR_MODULE_RDF         "rdf"         \
    $NS_ERROR_MODULE_UCONV       "uconv"       \
    $NS_ERROR_MODULE_REG         "registry"    \
    $NS_ERROR_MODULE_FILES       "files"       \
    $NS_ERROR_MODULE_DOM         "dom"         \
    $NS_ERROR_MODULE_IMGLIB      "imglib"      \
    $NS_ERROR_MODULE_MAILNEWS    "mailnews"    \
    $NS_ERROR_MODULE_EDITOR      "editor"      \
    $NS_ERROR_MODULE_XPCONNECT   "xpconnect"   \
    $NS_ERROR_MODULE_PROFILE     "profile"     \
    $NS_ERROR_MODULE_LDAP        "ldap"        \
    $NS_ERROR_MODULE_GENERAL     "general"     \
  ]

  array set errormsg [list \
    $NS_ERROR_NOT_INITIALIZED              "instance is not initialized"                   \
    $NS_ERROR_ALREADY_INITIALIZED          "instance is already initialized"               \
    $NS_ERROR_NOT_IMPLEMENTED              "function is not implemented"                   \
    $NS_ERROR_NO_INTERFACE                 "interface is not supported"                    \
    $NS_ERROR_NULL_POINTER                 "null pointer"                                  \
    $NS_ERROR_ABORT                        "function aborted"                              \
    $NS_ERROR_FAILURE                      "function failed"                               \
    $NS_ERROR_UNEXPECTED                   "unexpected error occured"                      \
    $NS_ERROR_OUT_OF_MEMORY                "out of memory"                                 \
    $NS_ERROR_INVALID_ARG                  "illegal value passed"                          \
    $NS_ERROR_NO_AGGREGATION               "class doesn't allow aggregation"               \
    $NS_ERROR_NOT_AVAILABLE                "resource unavailable"                          \
    $NS_ERROR_FACTORY_NOT_REGISTERED       "class is not registered"                       \
    $NS_ERROR_FACTORY_REGISTER_AGAIN       "class cannot be registered, try again later"   \
    $NS_ERROR_FACTORY_NOT_LOADED           "dynamically loaded factory could not be found" \
    $NS_ERROR_FACTORY_NO_SIGNATURE_SUPPORT "factory doesn't support signatures"            \
    $NS_ERROR_FACTORY_EXISTS               "factory is already registered"                 \
    $NS_ERROR_PROXY_INVALID_IN_PARAMETER   "could not create an in parameter proxy"        \
    $NS_ERROR_PROXY_INVALID_OUT_PARAMETER  "could not create an out parameter proxy"       \
  ]
}

###################################################
### xpcom::errormsg
### function called by the C library that returns
### an error message corresponding to the given
### XPCOM result code
###################################################
proc xpcom::errormsg {code} {
  variable errormsg
  variable modulenames

  set result "xpcom error: "

  if {[::info exists errormsg($code)]} {
    append result $errormsg($code)
  } else {
    set subcode      [expr {$code & 0x0000FFFF}]
    set modulecode   [expr {((($code >> 16) - $::xpcom::NS_ERROR_MODULE_BASE_OFFSET) & 0x1FFF)}] 

    if {[::info exists modulenames($modulecode)]} {
      set modulename $modulenames($modulecode)
    } else {
      set modulename $modulecode
    }
      
    if {[isSuccessCode $code]} {
      set severity "success"
    } else {
      set severity "failure"
    }

    append result "unknown (code: $subcode, module: $modulename, severity: $severity)"
  }

  set result
}


###################################################
### xpcom::wrap
### creates a new tcl component implementing the
### given list of interfaces with the given set of
### bindings
###################################################
#proc xpcom::wrap {args} {
#  ### XXX TODO: provide a wrapper for [xpcom::wrap] that accepts an optional argument
#  ### specifying which interface to return for the newly created component, similar
#  ### to [xpcom::new]
#}


###################################################
### xpcom::new
### creates a new component or service
###################################################
proc xpcom::new {args} {
  set index 0
  set service 0

  if {[string equal [lindex $args 0] "-service"]} {
    incr index
    set service 1
  }

  set argcount [expr {[llength $args] - $index}]
  if {$argcount < 1 || $argcount > 2} {
    error "wrong # args: should be \"?-service? classid ?interface?\""
  }

  set classid [lindex $args $index]

  if {$argcount == 2} {
    set interface [lindex $args [expr {$index + 1}]]
  } else {
    set interface nsISupports
  }

  if {$service} {
    return [$xpcom::smanager getService $classid $interface]
  } else {
    return [$xpcom::compmanager createInstance $classid [xpcom::null] $interface]
  }
}


###################################################
### xpcom::weakref
### returns a weak reference wrapped around
### the given interface reference
###################################################
proc xpcom::weakref {ref} {
  [$ref QueryInterface nsISupportsWeakReference] GetWeakReference
}


###################################################
### xpcom::throw
### throws an xpcom exception
###################################################
proc xpcom::throw {code} {
  return -code error -errorcode [list XPCOM $code] "Unhandled XPCOM Error (code: $code)"
}


###################################################
### xpcom::isSuccessCode
### returns 1 if the given XPCOM result code
### represents a successful return
###################################################
proc xpcom::isSuccessCode {code} {
  expr {!(($code >> 31) & 0x1)}
}


###################################################
### xpcom::lastResult
### returns the result code of the last
### xpcom method invocation
###################################################
proc xpcom::lastResult {} {
  set xpcom::lastResult
}


###################################################
### xpcom::interfaces
### returns a list of all the known interfaces
###################################################
proc xpcom::interfaces {} {
  set result {}

  foreach interface [getEnumeration [xpcom::invoke -noscript $xpcom::infomanager enumerateInterfaces] nsIInterfaceInfo] {
    lappend result [xpcom::invoke -noscript $interface set name]
  }

  set result
}


###################################################
### xpcom::classes
### returns a list of all the known class names
###################################################
proc xpcom::classes {} {
  set result {}

  set registrar [$xpcom::compmanager QueryInterface nsIComponentRegistrar]

  foreach cid [getSimpleEnumeration [$registrar enumerateContractIDs] nsISupportsString] {
    lappend result [$cid set data]
  }

  set result
}


###################################################
### xpcom::cids
### returns a list of all the known classids
###################################################
proc xpcom::cids {} {
  set result {}

  set registrar [$xpcom::compmanager QueryInterface nsIComponentRegistrar]

  foreach cid [getSimpleEnumeration [$registrar enumerateCIDs] nsISupportsID] {
    lappend result [$cid toString]
  }

  set result
}


###################################################
### xpcom::getEnumeration
### iterates through an nsIEnumeration, returns
### a list of references, optionally QId to a given
### interface
###################################################
proc xpcom::getEnumeration {enum {interface ""}} {
  set result {}

  $enum first

  while {![catch {$enum next}]} {
    set temp [$enum currentItem]
    if {$interface != ""} {set temp [$temp QueryInterface $interface]}
    lappend result $temp
  }

  set result
}


###################################################
### xpcom::getSimpleEnumeration
### iterates through an nsISimpleEnumeration, returns
### a list of references, optionally QId to a given
### interface
###################################################
proc xpcom::getSimpleEnumeration {enum {interface ""}} {
  set result {}

  while {[$enum hasMoreElements]} {
    set code [catch {$enum getNext} temp]
    if {$code} {
      break
    }
#    set temp [$enum getNext]
    if {$interface != ""} {set temp [$temp QueryInterface $interface]}
    lappend result $temp
  }

  set result
}


###################################################
### xpcom::bindings
### generates a binding list that maps a component's
### attributes and methods to a namespace's local
### variables and procedures
### for use as an input to xpcom::wrap
###################################################
proc xpcom::bindings {interfaces namespace} {
  set result {}

  set attributes [xpcom::info attributes $interfaces]
  set methods [xpcom::info methods $interfaces]

  foreach attribute $attributes {
    lappend result $attribute [list set ${namespace}::$attribute]
  }

  foreach method $methods {
    if {$method != "AddRef" && $method != "Release" && $method != "QueryInterface"} {
      lappend result $method [list ${namespace}::$method]
    }
  }

  set result
}


#######################################################
### xpcom::prototype
### generates a prototype implementation of a set
### of interfaces, encapsulated within a namespace
#######################################################
proc xpcom::prototype {interfaces namespace} {
  set result ""
  append result "namespace eval $namespace \{\n"
  set methods [xpcom::info methods $interfaces]
  set attributes [xpcom::info attributes $interfaces]

  append result "\n"
  append result "### Attributes ###\n"

  foreach attribute $attributes {
    append result [format "%-30s%s\n" "  variable $attribute" "  ; ## [xpcom::info type $interfaces $attribute]"]
  }

  append result "\n"
  append result "### Methods ###\n"
  
  foreach method $methods {
    catch {unset argnames}
    set params {}
    set outparams {}

    set retval [xpcom::info retval $interfaces $method]
    if {$retval != ""} {
      append result "  ## Returns $retval\n"
    }

    set arguments [xpcom::info args $interfaces $method]
    foreach argument $arguments {
      foreach {direction type} $argument {}
      set argname [join $type "_"]

      if {[::info exists argnames($argname)]} {
        incr argnames($argname)
        append argname "_$argnames($argname)"
      } else {
        set argnames($argname) 1
      }

      if {![string equal $direction "in"]} {
        lappend outparams [list $argname "${argname}_${direction}"]
        append argname "_$direction"
      }

      lappend params $argname
    }

    append result "  proc $method \{"
    append result [join $params " "]
    append result "\} \{\n"

    if {[llength $outparams]} {
      append result "    upvar"

      foreach outparam $outparams {
        append result " \$[lindex $outparam 1] [lindex $outparam 0]"
      }

      append result "\n"
    }

    append result "  \}\n\n"
  }

  append result "\}\n"

  append result "set tclcomponent_$namespace \[xpcom::wrap \{\n"
  foreach interface $interfaces {
    append result "  $interface\n"
  }

  append result "\} \[xpcom::bindings \{\n"
  foreach interface $interfaces {
    append result "  $interface\n"
  }

  append result "\} $namespace\]\]\n"
 
  set result
}


#######################################################
### xpcom::tostring
### returns the string representation of an interface ref
#######################################################
proc xpcom::tostring {reference} {
  format %s $reference
}


#######################################################
### xpcom::duplicate
### duplicates an interfaceref
#######################################################
proc xpcom::duplicate {interfaceref} {
  xpcom::fromstring [xpcom::tostring $interfaceref]
}


#######################################################
### xpcom::idequal
### compares two ids
#######################################################
proc xpcom::idequal {ida idb} {
  string equal [xpcom::id $ida] [xpcom::id $idb]
}


#######################################################
### xpcom::QueryInterface
### default implementation of QueryInterface for
### Tcl components
#######################################################
proc xpcom::QueryInterface {classinfo iid} {
  if {[xpcom::idequal $iid nsIClassInfo]} {
    if {![xpcom::isnull $classinfo]} {
      return $classinfo
    } else {
      xpcom::throw $xpcom::NS_ERROR_NO_INTERFACE
    }
  } else {
    return [xpcom::invoke -noscript [xpcom::this] QueryInterface $iid]
  }
}


#######################################################
### xpcom::factory_createinstance
### this is called by an implemented factory to
### actually create a tcl component
### it calls the binding script that was provided to
### xpcom::factory and automatically modifies the
### given binding list to support aggregation
### this list is then fed to xpcom::wrap in order
### to generate the component
#######################################################
proc xpcom::factory_createinstance {interfaces bindingscript classinfo outer iid} {
  array set bindings [eval $bindingscript]
  set bindings(QueryInterface) [list xpcom::QueryInterface $classinfo]

  if {![xpcom::isnull $outer]} {
    #### TODO: override QueryInterface for aggregation if outer is specified
    xpcom::throw $xpcom::NS_ERROR_NO_AGGREGATION
  }

  set result [xpcom::wrap -noscript -hidden -notxpcom $interfaces [array get bindings]]

  $result QueryInterface $iid
}


#######################################################
### xpcom::factory
### returns a new factory object that can be used to
### create instances of the component defined by the
### given interface list and binding script
### the binding script must return a proper binding
### list for a new component
#######################################################
proc xpcom::factory {interfaces bindingscript classinfo} {
  set bindings(createInstance) [list xpcom::factory_createinstance $interfaces $bindingscript $classinfo]
#  set bindings(lockFactory)    {???}

  set result [xpcom::wrap nsIFactory [array get bindings]]
  $result QueryInterface nsIFactory
}


#######################################################
### xpcom::module_getclassobject
#######################################################
proc xpcom::module_getclassobject {interp cmanager classid iid} {
  array set classes [$interp eval [list array get xpcom::classes]]
  foreach clsid [array names classes] {
    if {[xpcom::idequal $classid $clsid]} {
      foreach {description contractid factory} $classes($clsid) {break}
      return [$factory QueryInterface $iid]
    }
  }
  xpcom::throw $xpcom::NS_ERROR_FACTORY_NOT_REGISTERED
}


#######################################################
### xpcom::module_registerself
#######################################################
proc xpcom::module_registerself {interp cmanager location loaderstring type} {
  array set classes [$interp eval [list array get xpcom::classes]]
  foreach clsid [array names classes] {
    foreach {description contractid factory} $classes($clsid) {break}
    puts "Registering Tcl Component: $clsid"
    $cmanager nsIComponentRegistrar::registerFactoryLocation $clsid $description $contractid $location $loaderstring $type
  }
  catch {set registerproc [$interp eval [list set xpcom::registerprocs(registerproc)]]}
  if {[::info exists registerproc]} {
    $interp eval [concat $registerproc [list $cmanager $location $loaderstring $type]]
  }
}


#######################################################
### xpcom::module_unregisterself
#######################################################
proc xpcom::module_unregisterself {interp cmanager location loaderstring} {
  catch {set unregisterproc [$interp eval [list set xpcom::registerprocs(unregisterproc)]]}
  if {[::info exists unregisterproc]} {
    $interp eval [concat $unregisterproc [list $cmanager $location $loaderstring]]
  }
  array set classes [$interp eval [list array get xpcom::classes]]
  foreach clsid [array names classes] {
    foreach {description contractid factory} $classes($clsid) {break}
    puts "Unregistering Tcl Component: $clsid"
    catch {$cmanager unregisterFactoryLocation $clsid $location}
  }
}


#######################################################
### xpcom::module_canunload
#######################################################
proc xpcom::module_canunload {interp cmanager} {
  return yes
}


#######################################################
### xpcom::module
#######################################################
proc xpcom::module {script {destructor ""}} {
  set interp [interp create]
  set code [catch {$interp eval $script} result]
  if {$code} {
    interp delete $interp
    return -code $code $result
  }

  set bindings(getClassObject)  [list xpcom::module_getclassobject $interp]
  set bindings(registerSelf)    [list xpcom::module_registerself $interp]
  set bindings(unregisterSelf)  [list xpcom::module_unregisterself $interp]
  set bindings(canUnload)       [list xpcom::module_canunload $interp]
  set bindings()                [list catch [list interp delete $interp]]

  append bindings() "\n$destructor"

  set result [xpcom::wrap nsIModule [array get bindings]]
  $result QueryInterface nsIModule
}


#######################################################
### xpcom::register
#######################################################
proc xpcom::register {clsid description contractid interfaces bindingscript} {
  variable classes

  if {[::info exists classes($clsid)]} {
    error "class $clsid is already registered"
  }

  set classinfo       [xpcom::classinfo $clsid $description $contractid $interfaces]
  set factory         [xpcom::factory $interfaces $bindingscript $classinfo]
  set classes($clsid) [list $description $contractid $factory]
}


#######################################################
### xpcom::unregister
#######################################################
proc xpcom::unregister {clsid} {
  variable classes

  catch {unset classes($clsid)}
}


#######################################################
### xpcom::registermodule
#######################################################
proc xpcom::registermodule {registerproc unregisterproc} {
  variable registerprocs

  set registerprocs(registerproc)   $registerproc
  set registerprocs(unregisterproc) $unregisterproc
}


#######################################################
### xpcom::classinfo_gethelperforlanguage
#######################################################
proc xpcom::classinfo_gethelperforlanguage {language} {
  xpcom::null
}


#######################################################
### xpcom::classinfo
### creates an nsIClassInfo object
#######################################################
proc xpcom::classinfo {clsid description contractid interfaces} {
  set bindings(contractID)              [list set {} $contractid]
  set bindings(classDescription)        [list set {} $description]
  set bindings(classID)                 [list set {} $clsid]
  set bindings(implementationLanguage)  [list set {} [xpcom::constant nsIProgrammingLanguage UNKNOWN]]
  set bindings(flags)                   [list set {} 0]
  set bindings(getInterfaces)           [list set {} $interfaces]
  set bindings(getHelperForLanguage)    [list xpcom::classinfo_gethelperforlanguage]

  set result [xpcom::wrap nsIClassInfo [array get bindings]]
  $result QueryInterface nsIClassInfo
}


set xpcom::smanager    [xpcom::info servicemanager]
set xpcom::compmanager [$xpcom::smanager QueryInterface nsIComponentManager]
set xpcom::infomanager [xpcom::new -service "@mozilla.org/xpti/interfaceinfomanager-service;1" nsIInterfaceInfoManager]
