package require tclxpcom

if {$tcl_platform(platform) == "unix"} {
  package require Gtk
  gtk::gtk init argv
  set appconfig(console) no
  set appconfig(favorites) no
  set appconfig(framecontainer) yes
} else {
  set appconfig(console) yes
  set appconfig(favorites) yes
  set appconfig(framecontainer) no
}

set apptitle "Mefzilla"

if {$appconfig(console)} {
  console title "$apptitle Console (Tcl v[info patchlevel])"
}


set homepage "http://www.google.com"
set favoritesdir {c:\windows\favorites}

frame .view
frame .view.left  -takefocus 1
frame .view.geckoframe -container $appconfig(framecontainer)
frame .view.right -takefocus 1

pack .view.right
pack .view.geckoframe -side left -fill both -expand true
pack .view.left  -side left

set geckoframe .view.geckoframe

frame .location


set notimplemented {error "not implemented"}

set menubar .menubar
menu $menubar -tearoff 0

set filemenu $menubar.filemenu
menu $filemenu -tearoff 0

set gomenu $menubar.gomenu
menu $gomenu -tearoff 0

set scriptmenu $menubar.scriptmenu
menu $scriptmenu -tearoff 0

$filemenu add command -label "New"        -command $notimplemented
$filemenu add command -label "Open..."    -command $notimplemented
$filemenu add command -label "Save"       -command $notimplemented
$filemenu add command -label "Save As..." -command $notimplemented
$filemenu add separator
$filemenu add command -label "Print Setup..." -command $notimplemented
$filemenu add command -label "Print..."       -command $notimplemented
$filemenu add separator
$filemenu add command -label "Close" -command {exit}

$gomenu add command -label "Back"    -command goBack
$gomenu add command -label "Forward" -command goForward
$gomenu add command -label "Home"    -command goHome

$scriptmenu add command -label "Show Console" -command {console show}

$menubar add cascade -label "File"   -menu $filemenu   -underline 0
$menubar add cascade -label "Go"     -menu $gomenu     -underline 0

if {$appconfig(console)} {
  $menubar add cascade -label "Script" -menu $scriptmenu -underline 0
}

. configure -menu $menubar


entry .location.entry -textvariable url
bind .location.entry <Return> {loadURL $url}
bind .location.entry <Button-1> {focus -force .location.entry}

button .location.back    -text "<-" -command goBack
button .location.forward -text "->" -command goForward
button .location.stop    -text "<>" -command stop
button .location.reload  -text "~"  -command reload
button .location.home    -text "%"  -command goHome

canvas .location.progress -width 100 -height 20
set progresstext [.location.progress create text \
                                            [expr {[.location.progress cget -width] / 2}] \
                                            [expr {[.location.progress cget -height] / 2}] \
                                            -text "" -state hidden]

pack .location.back     -side left
pack .location.forward  -side left
pack .location.stop     -side left
pack .location.reload   -side left
pack .location.home     -side left
pack .location.entry    -side left -fill x -expand true
pack .location.progress -side left


label .status -textvariable status -anchor w

pack .location -fill x
pack .view -fill both -expand true
pack .status -fill x

wm geometry . =500x500

proc loadURL {url} {
  global webnav
  $webnav loadURI $url \
                  [xpcom::constant nsIWebNavigation LOAD_FLAGS_NONE] \
                  [xpcom::null] \
                  [xpcom::null] \
                  [xpcom::null]
}

proc goHome {} {
  global homepage
  loadURL $homepage
}

proc goBack {} {
  global webnav
  catch {$webnav goBack}
}

proc goForward {} {
  global webnav
  catch {$webnav goForward}
}

proc reload {} {
  global webnav
  $webnav reload [xpcom::constant nsIWebNavigation LOAD_FLAGS_NONE]
}

proc stop {} {
  global webnav
  $webnav stop [xpcom::constant nsIWebNavigation STOP_ALL]
}


proc findFile {pattern {directory .} {recursive no} {fullpath no}} {
  set filelist ""
  set pwd [pwd]

  if [catch {cd $directory} err] {
    return
  }

  foreach match [eval glob -nocomplain -- $pattern] {
    if {$fullpath} {
      lappend filelist [file join $directory $match]
    } else {
      lappend filelist $match
    }
  }

  if {$recursive} {
    foreach file [glob -nocomplain *] {
      if [file isdirectory $file] {
        set filelist [concat $filelist [findFile [file join $directory $file] $pattern $recursive $fullpath]]
      }
    }
  }

  cd $pwd
  return $filelist
}

proc readFavorites {dir} {
  set folders   {}
  set favorites {}
  foreach file [findFile * $dir no yes] {
    if [file isdirectory $file] {
      lappend folders $file
    } else {
      lappend favorites $file
    }
  }
  return [list $folders $favorites]
}

proc createFavoritesMenu {dir id} {
  set count 0
  menu $id -tearoff 0
  foreach {folders favorites} [readFavorites $dir] {break}
  foreach folder [lsort $folders] {
    $id add cascade -label [file tail $folder] \
                    -menu  [createFavoritesMenu $folder $id.$count]
    incr count
  }
  foreach favorite [lsort $favorites] {
    $id add command -label   [file rootname [file tail $favorite]] \
                    -command [list goToFavorite $favorite]
  }
  return $id
}

proc readFile {filename} {
  set ofile [open $filename r]
  set lines {}
  while {![eof $ofile]} {
    lappend lines [gets $ofile]
  }
  close $ofile
  return $lines
}


proc readFavorite {favorite} {
  set lines [readFile $favorite]
  set index [lsearch -glob $lines "URL=*"]
  if {$index != -1} {
    return [string range [lindex $lines $index] 4 end]
  } else {
    return ""
  }
}

proc goToFavorite {favorite} {
  set url [readFavorite $favorite]
  if {$url != ""} {
    loadURL $url
  }
}

if {$appconfig(favorites)} {
  $menubar add cascade -label "Favorites" \
                     -menu [createFavoritesMenu $favoritesdir $menubar.favorites]
}

proc map {func list} {
  set result {}
  foreach item $list {
    lappend result [eval [concat $func $item]]
  }
  return $result
}

proc kilobyteDesc {bytes} {
  if {$bytes > 0} {
    return [expr {round($bytes / 1000)}]K
  } else {
    return ""
  }
}

proc drawprogressbar {widget curtotal maxtotal} {
  global progresstext
  $widget itemconfigure $progresstext -state normal \
             -text [join [eval concat [map kilobyteDesc [list $curtotal $maxtotal]]] "/"]
}

proc getSecurityLevelInfo {securityflags} {
  set result {}

  if {$securityflags & [xpcom::constant nsIWebProgressListener STATE_SECURE_LOW]} {
    set result "LOW"
  } elseif {$securityflags & [xpcom::constant nsIWebProgressListener STATE_SECURE_MED]} {
    set result "MEDIUM"
  } else {
    set result "HIGH"
  }

  set result
}

proc getSecurityStateInfo {securityflags} {
  set result {}

  if {$securityflags & [xpcom::constant nsIWebProgressListener STATE_IS_SECURE]} {
    set result "SECURE"
  } elseif {$securityflags & [xpcom::constant nsIWebProgressListener STATE_IS_BROKEN]} {
    set result "BROKEN"
  } else {
    set result "INSECURE"
  }

  set result
}

proc getProgressTypeInfo {stateflags} {
  set result {}

  if {$stateflags & [xpcom::constant nsIWebProgressListener STATE_IS_REQUEST]} {
    set result "REQUEST"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_IS_DOCUMENT]} {
    set result "DOCUMENT"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_IS_NETWORK]} {
    set result "NETWORK"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_IS_WINDOW]} {
    set result "WINDOW"
  }

  set result
}

proc getProgressStateInfo {stateflags} {
  set result {}

  if {$stateflags & [xpcom::constant nsIWebProgressListener STATE_START]} {
    set result "START"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_REDIRECTING]} {
    set result "REDIRECTING"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_TRANSFERRING]} {
    set result "TRANSFERRING"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_NEGOTIATING]} {
    set result "NEGOTIATING"
  } elseif {$stateflags & [xpcom::constant nsIWebProgressListener STATE_STOP]} {
    set result "STOP"
  }

  set result
}

proc dumpSecurityInfo {securityflags} {
  puts "Security State: [getSecurityStateInfo $securityflags]"
  puts "Security Level: [getSecurityLevelInfo $securityflags]"
}

proc dumpProgressInfo {stateflags} {
  puts "Progress Type:  [getProgressTypeInfo  $stateflags]"
  puts "Progress State: [getProgressStateInfo $stateflags]"
}

proc dumpLoadGroupInfo {loadgroup} {
  puts "LoadGroup:   $loadgroup"
  if {![xpcom::isnull $loadgroup]} {
    puts "LoadGroup activecount: [$loadgroup set activeCount]"
    puts "LoadGroup requests: [xpcom::getSimpleEnumeration [$loadgroup set requests]]"
  }
}

proc dumpRequestInfo {request} {
  if {[xpcom::isnull $request]} {
    return
  }

  puts "Request url: [$request set name]"

  ## isPending may not be implemented, causing an XPCOM 'exception'
  puts "IsPending:   [$request isPending]"
  puts "Status:      [$request set status]"

  dumpLoadGroupInfo [$request set loadGroup]
}

proc dumpWebProgress {webprogress} {
  puts "Web Progress: $webprogress"
  puts "DOM Window:   [$webprogress set DOMWindow]"
}

proc titlechange {window varname1 varname2 op} {
  global apptitle
  upvar $varname1 title
  if {$op == "w"} {
    wm title $window "$title - $apptitle"
  }
}


proc initMozilla {} {
  global eventqueue profmanager
  set eqservice [xpcom::new -service "@mozilla.org/event-queue-service;1" nsIEventQueueService]
  $eqservice createThreadEventQueue

  set eventqueue [$eqservice getSpecialEventQueue [xpcom::constant nsIEventQueueService CURRENT_THREAD_EVENT_QUEUE]]

  set profmanager [xpcom::new -service "@mozilla.org/profile/manager;1" nsIProfileInternal]
  xpcom::invoke -hidden $profmanager startupWithArgs [xpcom::null] yes
}

proc initBrowser {chrome} {
  global browser basewindow sitewindow webnav

  set browser    [xpcom::new "@mozilla.org/embedding/browser/nsWebBrowser;1" nsIWebBrowser]
  set basewindow [$browser QueryInterface nsIBaseWindow]
  set webnav     [$browser QueryInterface nsIWebNavigation]

  $browser set containerWindow $chrome
  $browser set nsIDocShellTreeItem::itemType \
               [xpcom::constant nsIDocShellTreeItem typeContentWrapper]

  $chrome nsIEmbeddingSiteWindow::getDimensions 0 x x width height
  set sitewindow [xpcom::invoke -hidden $chrome set nsIEmbeddingSiteWindow::siteWindow]
  xpcom::invoke -hidden $basewindow initWindow $sitewindow [xpcom::null] 0 0 $width $height

  $basewindow create
  $chrome set webBrowser $browser
  $basewindow set visibility true

  $browser addWebBrowserListener [xpcom::weakref $chrome] nsIWebProgressListener
  $browser nsIWebBrowserFocus::activate
}


namespace eval ::chrome {
  variable view                ; ## tk frame object
  variable visibility   true   ; ## boolean
  variable title               ; ## wstring
  variable siteWindow          ; ## pointer
  variable webBrowser          ; ## nsIWebBrowser
  variable chromeFlags  0      ; ## uint32

  variable interfaces {
    nsIWebBrowserChrome
    nsIEmbeddingSiteWindow
    nsIInterfaceRequestor
    nsIWebProgressListener
    nsIContextMenuListener
    nsIWebBrowserChromeFocus
  }

  proc init {_view} {
    global tcl_platform
    variable view
    variable siteWindow
    variable interfaces
    set view $_view

    if {$tcl_platform(platform) == "unix"} {
      set w [gtk::plug new [winfo id $view]]
      gtk::widget show $w
      set siteWindow [xpcom::pointer $w]
    } else {
      set siteWindow [xpcom::pointer [winfo id $view]]
    }

    set chrome [xpcom::wrap -noscript -hidden $interfaces [xpcom::bindings $interfaces ::chrome]]
    return [$chrome QueryInterface nsIWebBrowserChrome]
  }

### Methods ###
  proc onStateChange {webprogress request flags status} {
#    dumpProgressInfo $flags
#    dumpWebProgress  $webprogress
#    dumpRequestInfo  $request
#    puts "status: [xpcom::errormsg $status]"
  }

  proc onProgressChange {webprogress request curself maxself curtotal maxtotal} {
#    dumpWebProgress $webprogress
#    dumpRequestInfo $request
    drawprogressbar .location.progress $curtotal $maxtotal
  }

  proc onLocationChange {webprogress request uri} {
    global url
#    dumpWebProgress $webprogress
#    dumpRequestInfo $request
    set url [$uri set spec]
  }

  proc onStatusChange {webprogress request newstatus message} {
    global status
#    dumpWebProgress $webprogress
#    dumpRequestInfo $request
    set status $message
  }

  proc onSecurityChange {webprogress request state} {
#    puts "onSecurityChange called: $webprogress, $request, $state"
#    dumpWebProgress $webprogress
#    dumpRequestInfo $request
#    dumpSecurityInfo $state
  }

  proc setDimensions {flags x y cx cy} {
    puts "setDimensions called: $flags, $x, $y, $cx, $cy"
  }

  proc getDimensions {flags outx outy outcx outcy} {
    variable view
    upvar $outx x $outy y $outcx cx $outcy cy
    set x  0
    set y  0
    set cx [$view cget -width]
    set cy [$view cget -height]
  }

  proc setFocus {} {
    global basewindow
    $basewindow setFocus
  }

  proc onShowContextMenu {flags event node} {
    puts "OnShowContextMenu called: $flags, $event, $node"
  }

  proc getInterface {iid} {
    global browser
    if {$iid == "nsIDOMWindow"} {
      return [$browser set contentDOMWindow]
    } else {
      return [[xpcom::this] QueryInterface $iid]
    }
  }

  proc setStatus {code newstatus} {
    global status
    set status $newstatus
  }

  proc destroyBrowserWindow {} {
    puts "destroyBrowserWindow called"
  }

  proc sizeBrowserTo {cx cy} {
    puts "sizeBrowserTo called: $cx, $cy"
  }

  proc showAsModal {} {
    puts "showAsModal called"
  }

  #returns boolean
  proc isWindowModal {} {
    return no
  }

  proc exitModalEventLoop {status} {
    puts "exitModalEventLoop called"
  }

  proc focusNextElement {} {
    focus -force [tk_focusNext .view.right]
  }

  proc focusPrevElement {} {
    focus -force [tk_focusPrev .view.left]
  }
}

initMozilla
set chrome [::chrome::init $geckoframe]
initBrowser $chrome
trace variable ::chrome::title w [list titlechange .]

bind $::chrome::view <Configure> {
  $basewindow setPositionAndSize 0 0 %w %h true
}

bind .view.right <FocusIn> {
  $browser nsIWebBrowserFocus::setFocusAtLastElement;  # currently does nothing!
  $browser nsIWebBrowserFocus::activate
}

bind .view.left <FocusIn> {
  $browser nsIWebBrowserFocus::setFocusAtFirstElement; # currently does nothing!
  $browser nsIWebBrowserFocus::activate
}

goHome

if {$tcl_platform(platform) == "unix"} {
  while 1 {
    $eventqueue processPendingEvents
    update
  }
}

